#include "metacraneclient.h"
#include <QJsonObject>
#include <QJsonDocument>
#include <QDesktopServices>
#include <QUrl>
#include <QTimer>
#include <QByteArray>
#include <QFile>
#include <QFileInfo>

const size_t BUFFER_SIZE = 250000;

#define METACRANE_DBUS_SERVICE "harbour.metacrane.client"
#define METACRANE_DBUS_PATH "/"
#define METACRANE_DBUS_INTERFACE "harbour.metacrane.client"

MetacraneClient::MetacraneClient(QObject *parent) :
    QObject(parent)
//,   m_iface(NULL)
,   m_next_transfer_id(1)
,   m_current_file(NULL)
,   m_current_file_stream(NULL)
,   m_current_transfer_id(-1)
,   m_current_transfer_num_pieces(0)
,   m_current_transfer_piece_index(0)
{
    if(!QDBusConnection::sessionBus().registerObject(METACRANE_DBUS_PATH, this,
            QDBusConnection::ExportScriptableContents)){
        qDebug()<<"WARNING: Failed to register DBus object";
    } else if(!QDBusConnection::sessionBus().registerService(METACRANE_DBUS_SERVICE)){
        qDebug()<<"WARNING: Failed to register DBus service";
    } else {
        qDebug()<<"DBus service OK.";
        /*m_iface = new QDBusInterface(METACRANE_DBUS_SERVICE, METACRANE_DBUS_PATH,
                METACRANE_DBUS_INTERFACE, QDBusConnection::sessionBus(), 0);
        m_iface->connection().connect(METACRANE_DBUS_SERVICE, METACRANE_DBUS_PATH,
                METACRANE_DBUS_INTERFACE, "sendFile", this,
                SLOT(messageSlot(QString,QString,QString,int)));*/
        /*QDBusConnection::sessionBus().connect(QString(), QString(), METACRANE_DBUS_INTERFACE,
                "sendFile", this, SLOT(messageSlot(QString,QString,QString,int)));*/
    }

    QObject::connect(&m_socket, SIGNAL(connected()), this, SLOT(socketConnected()));
    QObject::connect(&m_socket, SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    addBuiltinHandlers();
    m_socket.pokeConnection();
    QTimer *timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(update()));
    timer->start(5000);
}

MetacraneClient::~MetacraneClient()
{
    delete m_current_file_stream;
    delete m_current_file;
}

void MetacraneClient::pokeConnection()
{
    m_socket.pokeConnection();
}

void MetacraneClient::gotoCrane(QString name, std::function<void(bool success)> cb)
{
    if(!m_socket.isConnected()){
        qDebug()<<"gotoCrane:"<<name<<": delaying until connected";
        pokeConnection();
        m_current_crane_name = name;
        currentCraneNameChanged();
        return;
    }
    qDebug()<<"gotoCrane:"<<name;
    QVariantMap params;
    params["name"] = name;
    sendRequest("join_crane", params,
            [this, name, cb](const QVariantMap &response)
    {
        qDebug()<<"join_crane response"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            cb(false);
            return;
        }
        m_current_crane_name = name;
        currentCraneNameChanged();
        cb(true);
    });
}

void MetacraneClient::gotoCrane(QString name)
{
    return gotoCrane(name, [](bool success){(void)success;});
}

void MetacraneClient::openUrl(QString url)
{
    QDesktopServices::openUrl(QUrl(url));
}

void MetacraneClient::clearMessage()
{
    m_last_message.clear();
    lastMessageChanged();
}

void MetacraneClient::uploadText(QString text, int deletion_time)
{
    QVariantMap params;
    params["text"] = text;
    params["confirm_downloads"] = false;
    params["deletion_time"] = deletion_time;
    sendRequest("lift_text", params,
            [this](const QVariantMap &response)
    {
        qDebug()<<"lift_text response"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            return;
        }
    });
}

void MetacraneClient::uploadFiles(QStringList paths, int deletion_time)
{
    qDebug()<<"uploadFiles:"<<paths;
    m_file_queue.append(paths);
    queueSizeChanged();
    // TODO: Queue the correct time for each file (so that multiple calls to
    //       this function with different times will lead to correct behavior)
    m_deletion_time = deletion_time;
    uploadNextFile();
}

void MetacraneClient::updateSavedCraneNames(QStringList crane_names)
{
    m_saved_crane_names = crane_names;
}

void MetacraneClient::addMessageHandler(QString name,MetacraneSocket::msg_cb_t handler)
{
    m_socket.addMessageHandler(name, handler);
}

double MetacraneClient::uploadProgress()
{
    if(m_current_file_stream == NULL || m_current_file == NULL)
        return -1;
    return (double)m_current_transfer_piece_index / m_current_transfer_num_pieces;
}

int MetacraneClient::queueSize()
{
    return m_file_queue.size();
}

// TODO: Excplicit mime type
void MetacraneClient::sendFile(QString path, QString fileName, QString craneName, int deletionTime)
{
    qDebug()<<"sendFile called; path:"<<path<<"fileName:"<<fileName
            <<"craneName:"<<craneName<<"deletionTime:"<<deletionTime;
    auto start_upload = [this, path, deletionTime](){
        QStringList paths;
        paths.append(path);
        m_deletion_time = deletionTime; // NOTE: This sets the global deletion time
        uploadFiles(paths, 56*3600);
    };
    if(m_current_crane_name == craneName){
        start_upload();
    } else {
        gotoCrane(craneName, [start_upload](bool success){
            if(success)
                start_upload();
            else
                qDebug()<<"Can't upload; crane switch failed";
        });
    }
}

QStringList MetacraneClient::getSavedCraneNames()
{
    return m_saved_crane_names;
}

void MetacraneClient::addBuiltinHandlers()
{
    addMessageHandler("message", [this](const QVariantMap &response){
        QString text = response.value("text", QString()).toString();
        qDebug()<<"message"<<text;
        m_last_message = "Server says: "+text;
        lastMessageChanged();
    });
    addMessageHandler("add_entity", [this](const QVariantMap &response){
        int entity = response.value("entity", -1).toInt();
        qDebug()<<"add_entity"<<entity;
        QVariantMap components = response.value("components", QVariant()).toMap();
        Entity *e = new Entity(entity);
        e->components() = components;
        m_entities.prepend(e);
        entitiesChanged();
    });
    addMessageHandler("delete_entity", [this](const QVariantMap &response){
        int entity = response.value("entity", -1).toInt();
        qDebug()<<"delete_entity"<<entity;
        QMutableListIterator<QObject*> i(m_entities);
        while (i.hasNext()) {
            Entity *e = (Entity*)i.next();
            if(e->entity() == entity)
                i.remove();
        }
        entitiesChanged();
    });
    // TODO: add_component
    // TODO: delete_component
}

void MetacraneClient::sendMessage(QString name, const QVariantMap &params)
{
    m_socket.sendMessage(name, params);
}

void MetacraneClient::sendRequest(QString name, const QVariantMap &params,
        std::function<void(const QVariantMap &response)> cb)
{
    m_socket.sendRequest(name, params, cb);
}

void MetacraneClient::showMessage(QString message)
{
    m_last_message = message;
    lastMessageChanged();
}

void MetacraneClient::uploadNextFile()
{
    if(m_current_file != NULL || m_current_file_stream != NULL){
        qDebug()<<"uploadNextFile: already uploading";
        return;
    }
    if(m_file_queue.empty()){
        qDebug()<<"uploadNextFile: queue empty";
        return;
    }
    QString path = m_file_queue.first();
    m_file_queue.pop_front();
    queueSizeChanged();
    qDebug()<<"uploadNextFile: "<<path;

    m_current_file = new QFile(path);
    QFile &f = *m_current_file;
    if(!f.open(QIODevice::ReadOnly)){
        showMessage(QString()+"Error: Could not open file for reading ("+path+")");
        endCurrentTransfer();
        return;
    }

    QFileInfo info(f);
    quint64 size = info.size();
    qDebug()<<"upload file size:"<<size;

    m_current_file_stream = new QDataStream(m_current_file);
    m_current_transfer_id = m_next_transfer_id++;
    m_current_transfer_num_pieces = ((size - 1) / BUFFER_SIZE) + 1;
    m_current_transfer_piece_index = 0;

    QVariantMap params;
    params["transfer_id"] = m_current_transfer_id;
    params["name"] = info.fileName();
    params["num_pieces"] = m_current_transfer_num_pieces;
    params["size"] = size;
    params["confirm_downloads"] = false;
    params["deletion_time"] = m_deletion_time;
    sendRequest("start_file_transfer", params,
            [this](const QVariantMap &response){
        qDebug()<<"start_file_transfer response:"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            endCurrentTransfer();
            return;
        }
        sendNextFilePiece();
    });
    uploadProgressChanged();
}

void MetacraneClient::sendNextFilePiece()
{
    if(m_current_file_stream == NULL || m_current_file == NULL)
        return;
    qDebug()<<"Sending piece"<<m_current_transfer_piece_index;
    qDebug()<<"m_file_stream"<<m_current_file_stream;
    QByteArray buf(BUFFER_SIZE, 0);
    int read_len = m_current_file_stream->readRawData(buf.data(), BUFFER_SIZE);
    qDebug()<<"read_len:"<<read_len;
    if(read_len != BUFFER_SIZE &&
            m_current_transfer_piece_index != m_current_transfer_num_pieces - 1){
        qDebug()<<"Read wrong size:"<<read_len;
        showMessage("Error: Failed to read full buffer");
        endCurrentTransfer();
        return;
    }
    QByteArray buf2(buf.data(), read_len);

    QVariantMap params;
    params["transfer_id"] = m_current_transfer_id;
    params["index"] = m_current_transfer_piece_index++;
    params["data_b64"] = buf2.toBase64();
    sendRequest("file_data", params,
            [this](const QVariantMap &response){
        qDebug()<<"file_data response:"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            endCurrentTransfer();
            return;
        }
        if(m_current_transfer_piece_index < m_current_transfer_num_pieces){
            sendNextFilePiece();
            return;
        }
        qDebug()<<"Transfer finished";
        endCurrentTransfer();
    });
    uploadProgressChanged();
}

void MetacraneClient::endCurrentTransfer()
{
    delete m_current_file_stream;
    m_current_file_stream = NULL;
    delete m_current_file;
    m_current_file = NULL;
    uploadProgressChanged();
    uploadNextFile();
}

void MetacraneClient::socketConnected()
{
    qDebug()<<"client connected";

    m_entities.clear();

    QVariantMap params;
    params["version"] = 2;
    sendMessage("version", params);

    if(m_current_crane_name != ""){
        qDebug()<<"joining delayed crane";
        gotoCrane(m_current_crane_name);
    }
}

void MetacraneClient::socketDisconnected()
{
    qDebug()<<"client disconnected";
}

void MetacraneClient::update()
{
    m_socket.pokeConnection();
}
