#include "metacraneclient.h"
#include <QJsonObject>
#include <QJsonDocument>
#include <QTimer>

MetacraneSocket::MetacraneSocket(QObject *parent) :
    QObject(parent),
    m_ws(NULL),
    m_conn_status(CS_DISCONNECTED),
    m_conn_status_timer(new QTimer(this)),
    m_next_req_id(1)
{
    m_ws = new QtWebsocket::QWsSocket(this, NULL);
    QObject::connect(m_ws, SIGNAL(stateChanged(QAbstractSocket::SocketState)), this, SLOT(socketStateChanged(QAbstractSocket::SocketState)));
    QObject::connect(m_ws, SIGNAL(frameReceived(QString)), this, SLOT(handleMessage(QString)));
    QObject::connect(m_ws, SIGNAL(connected()), this, SLOT(socketConnected()));
    QObject::connect(m_ws, SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    QObject::connect(m_ws, SIGNAL(sslErrors(const QList<QSslError>&)), this, SLOT(handleSslErrors(const QList<QSslError>&)));
    addBuiltinHandlers();
    // Intentionally non-single-shot
    QObject::connect(m_conn_status_timer, SIGNAL(timeout()), this, SLOT(connectionTimeout()));
}

void MetacraneSocket::pokeConnection()
{
    //qDebug()<<"pokeConnection";
    if(m_conn_status == CS_DISCONNECTED){
        QString host = "metacrane.com";
        quint16 port = 443;
        qDebug()<<"Connecting to"<<host<<"port"<<port;
        m_ws->connectToHost(host, port, "/websocket", true);
        m_conn_status = CS_CONNECTING;
        m_conn_status_timer->start(10000);
    }
}

bool MetacraneSocket::isConnected()
{
    return m_conn_status == CS_CONNECTED;
}

void MetacraneSocket::addMessageHandler(QString name, MetacraneSocket::msg_cb_t handler)
{
    m_msg_cbs[name] = handler;
}

void MetacraneSocket::addBuiltinHandlers()
{
    addMessageHandler("req_resp", [this](const QVariantMap &response){
        int req_id = response.value("req_id", -1).toInt();
        qDebug()<<"req_resp"<<req_id;
        auto it = m_req_cbs.find(req_id);
        if(it == m_req_cbs.end()){
            qDebug()<<"error: no handler for"<<req_id;
            return;
        }
        req_cb_t cb = it.value();
        QVariantMap params = response.value("data", QVariant()).toMap();
        cb(params);
        m_req_cbs.erase(it);
    });
}

void MetacraneSocket::connectionTimeout()
{
    if(m_conn_status == CS_CONNECTING){
        qDebug()<<"timeout while connecting; disconnected";
        m_ws->disconnect();
        m_conn_status = CS_DISCONNECTED;
    }
}

void MetacraneSocket::sendMessage(QString name, const QVariantMap &params)
{
    QVariantMap data;
    data["name"] = name;
    data["data"] = params;
    QJsonObject data_object = QJsonObject::fromVariantMap(data);
    QJsonDocument data_json(data_object);
    QByteArray ba = data_json.toJson();
    m_ws->write(ba);
}

void MetacraneSocket::sendRequest(QString name, const QVariantMap &params,
        std::function<void(const QVariantMap &response)> cb)
{
    int req_id = m_next_req_id++;
    QVariantMap data;
    data["name"] = name;
    data["data"] = params;
    data["req_id"] = req_id;
    sendMessage("req", data);
    m_req_cbs[req_id] = cb;
}

void MetacraneSocket::socketConnected()
{
    qDebug()<<"socket connected";
    m_conn_status = CS_CONNECTED;
    m_conn_status_timer->start(10000);
    connected();
}

void MetacraneSocket::socketDisconnected()
{
    qDebug()<<"socket disconnected";
    m_conn_status = CS_DISCONNECTED;
    m_conn_status_timer->start(10000);
    disconnected();
}

void MetacraneSocket::handleMessage(QString message)
{
    qDebug()<<"received:"<<message;
    QJsonDocument message_json = QJsonDocument::fromJson(message.toUtf8());
    QJsonObject message_object = message_json.object();
    QString name = message_object["name"].toString();
    auto it = m_msg_cbs.find(name);
    if(it == m_msg_cbs.end()){
        qDebug()<<"error: no handler for"<<name;
        return;
    }
    msg_cb_t cb = it.value();
    QJsonObject params_object = message_object["data"].toObject();
    QVariantMap params = params_object.toVariantMap();
    cb(params);
}

void MetacraneSocket::socketStateChanged(QAbstractSocket::SocketState socketState)
{
    qDebug()<<"socket state changed:"<<socketState;
}

void MetacraneSocket::handleSslErrors(const QList<QSslError> &errors)
{
    qDebug()<<"ssl errors:"<<errors;
}
